/**
 * Qt C++ Quad Tree
 *
 * Stores templated type with specific positions in a quad tree.
 * 
 * Ray Rischpater, Nokia Corporation
 *
 * Inspired by code elsewhere on the net, including
 * http://veeableful.wordpress.com/2012/02/07/algorithm-space-partitioning-using-quadtree/
 *
 * This file is in the public domain.
 */

#ifndef QUADTREE_H_
#define QUADTREE_H_
#include <QList>
#include <QPointF>
#include <QSizeF>
#include <QRectF>

#include <QDebug>

/**
 * A single node in the QuadTree.
 * Consists of a point and an element of type T.
*/
template <typename T>
class QuadTreeNode {
public:
    QuadTreeNode(const QPointF& location, const T& data): mLocation(location), mData(data) {}
    QPointF location() const { return mLocation; }
    double x() const { return mLocation.x(); }
    double y() const { return mLocation.y(); }
    T data() const   { return mData;         }
    QuadTreeNode& operator =(const QuadTreeNode& other) {
        if (this != &other) {
            this->mLocation = other.mLocation;
            this->mData = other.mData;
        }
        return *this;
    }

private:
    QPointF mLocation;
    T mData;
};

typedef QuadTreeNode<QVariant> VariantQuadTreeNode;

template <typename T>
class QuadTree
{
public:
    /** Constructor */
    QuadTree(const QPointF& location, const QSizeF& size, int level = 0, int maxLevel = 18);

    virtual ~QuadTree();

    /**
     * Add an object to the tree which will then be processed
     * by the quadtree. The quad tree will make a copy of the
     * object.
     */
    void add(const QuadTreeNode<T>& object);
    void add(const QPointF& where, const T& object);

    /**
     * Return objects in the given region.
     */
    QList< QuadTreeNode<T> > find(const QRectF& region) const;

private:
    /**
     * Add objects to the tree which will then be processed
     * by the quadtree. The quad tree now owns the object at
     * pObject.
     */
    void add(QuadTreeNode<T>* pObject);

    /** To pass all the objects in storage to child nodes */
    void split();

    /** Pass a single object to a child node. Figures out
     *  which child node should take the object, and creates
     *  the child node if necessary.
     */
    void pass(QuadTreeNode<T>* object);

    /** "Location" of the nodes */
    enum QuadTreeDirections
    {
        NW,
        NE,
        SW,
        SE,
        DirectionCount
    };


    /**
     * Location and Dimension are used to determine
     * which node an object should go into
     */
    QPointF mLocation;
    QSizeF mSize;

    /** Quadtree current depth and max depth */
    int mLevel;
    int mMaxLevel;


    /** Storage for objects */
    QList<QuadTreeNode<T>* > mObjects;

    /** Objects counter */
    unsigned int mObjectCount;

    /** The branching nodes */
    QuadTree* mNodes[DirectionCount];
};
typedef QuadTree<QVariant> VariantQuadTree;

/*
 * Implementation follows
 */

template<typename T>
QuadTree<T>::QuadTree(const QPointF& location, const QSizeF& size, int level, int maxLevel)
    : mLocation(location)
    , mSize(size)
    , mLevel(level)
    , mMaxLevel(maxLevel)
{
    mObjectCount = 0;   // NumObjects should start at 0
    for (int i = 0; i < DirectionCount; i++)
    {
        mNodes[i] = 0; // Initialize pointers to 0
    }
}

template<typename T>
QuadTree<T>::~QuadTree()
{
    // Free our data for this node
    while(!mObjects.isEmpty()) {
        QuadTreeNode<T>* pObj = mObjects.takeFirst();
        delete pObj;
    }

    // And free child nodes
    for (int i = 0; i < DirectionCount; i++) {
        delete mNodes[i];
    }

}

template<typename T>
void QuadTree<T>::add(const QuadTreeNode<T>& object)
{
    QuadTreeNode<T>* pObject = 0;
    // Make a heap copy of the object if we're top-of-tree, so we can own this object.
    pObject = new QuadTreeNode<T>(object.location(), object.data());

    add(pObject);
}

template<typename T>
void QuadTree<T>::add(QuadTreeNode<T>* pObject)
{
    // Maximum number of objects to hold before it splits into 4 nodes
    const unsigned int MAXOBJECTS = 3;
    ++mObjectCount;

    // Don't pass objects down to the nodes until we reach the threshold or
    // we're at the bottom of the tree, instead store it for now.
    if (mObjectCount <= MAXOBJECTS || mLevel == mMaxLevel) {
        mObjects.append(pObject);

        // Diverge the objects to child nodes if we reach the threshold
        // and we can go deeper
        if (mObjectCount == MAXOBJECTS && mLevel != mMaxLevel) {
            split();
        }
    } else {
        /** No longer store the objects, let child nodes store them */
        pass(pObject);
    }
}

template<typename T>
void QuadTree<T>::pass(QuadTreeNode<T>* pObj) {
    QSizeF newDimension(mSize.width()/2, mSize.height()/2);
    if (pObj->x() >= mLocation.x() + mSize.width()/2) {
        if (pObj->y() >= mLocation.y() + mSize.height()/2) {
            if (mNodes[SE] == 0) {
                QPointF newLocation(mLocation.x() + mSize.width()/2,
                                    mLocation.y() + mSize.height()/2);
                mNodes[SE] = new QuadTree<T>(newLocation, newDimension, ++mLevel, mMaxLevel);
            }
            mNodes[SE]->add(pObj);
        } else {
            if (mNodes[NE] == 0) {
                QPointF newLocation(mLocation.x() + mSize.width()/2, mLocation.y());
                mNodes[NE] = new QuadTree<T>(newLocation, newDimension, ++mLevel, mMaxLevel);
            }
            mNodes[NE]->add(pObj);
        }
    } else if (pObj->y() >= mLocation.y() + mSize.height()/2) {
        if (mNodes[SW] == 0)
        {
            QPointF newLocation(mLocation.x(), mLocation.y() + mSize.height()/2);
            mNodes[SW] = new QuadTree<T>(newLocation, newDimension, ++mLevel, mMaxLevel);
        }
        mNodes[SW]->add(pObj);
    } else {
        if (mNodes[NW] == 0)
        {
            QPointF newLocation(mLocation.x(), mLocation.y());
            mNodes[NW] = new QuadTree<T>(newLocation, newDimension, ++mLevel, mMaxLevel);
        }
        mNodes[NW]->add(pObj);
    }
}

template<typename T>
void QuadTree<T>::split()
{
    // Simply pass off each of these nodes to their child trees.
    while(!mObjects.isEmpty()) {
        QuadTreeNode<T>* pObj = mObjects.takeFirst();
        pass(pObj);
    }
}

template<typename T>
QList< QuadTreeNode<T> > QuadTree<T>::find(const QRectF& region) const
{
    QList< QuadTreeNode<T> > possibleResults;

    // Do we intersect? If so, return our elements.
    if (region.intersects(QRectF(mLocation, mSize))) {
        // Leaf node? Return the nodes here.
        if (mObjects.length()) {
            foreach(const QuadTreeNode<T>* pNode, mObjects)
                possibleResults.append(*pNode);
        } else {
        // Traverse our child trees.
            for(int i = 0; i < DirectionCount; i++) {
                // If there's a sub-tree here, traverse it.
                if (mNodes[i]) {
                    possibleResults.append(mNodes[i]->find(region));
                }
            }
        }

        // Do a filter on the returned objects,
        // returning only those that actually intersect the region.

        QMutableListIterator< QuadTreeNode<T> > iterator(possibleResults);
        while (iterator.hasNext()) {
            if (!region.contains(iterator.next().location()))
                iterator.remove();
        }
    }
    return possibleResults;
}

#endif
